@@ -12,14 +12,15 @@ use serde::{
         Deserialize, DeserializeSeed, Deserializer, Error, MapAccess, SeqAccess, Unexpected,
         Visitor,
     },
-    ser::{Serialize, SerializeSeq, SerializeStruct, SerializeTupleStruct, Serializer},
+    ser::{
+        Serialize, SerializeMap, SerializeSeq, SerializeStruct, SerializeTupleStruct, Serializer,
+    },
 };
 use static_assertions::assert_impl_all;
 
 use crate::{
-    array_display_fmt, dict_display_fmt, signature_parser::SignatureParser, structure_display_fmt,
-    utils::*, Array, Basic, Dict, DynamicType, ObjectPath, OwnedValue, Signature, Str, Structure,
-    StructureBuilder, Type,
+    array_display_fmt, dict_display_fmt, structure_display_fmt, utils::*, Array, Basic, Dict,
+    DynamicType, ObjectPath, OwnedValue, Signature, Str, Structure, StructureBuilder, Type,
 };
 #[cfg(feature = "gvariant")]
 use crate::{maybe_display_fmt, Maybe};
@@ -88,7 +89,7 @@ pub enum Value<'a> {
     U64(u64),
     F64(f64),
     Str(Str<'a>),
-    Signature(Signature<'a>),
+    Signature(Signature),
     ObjectPath(ObjectPath<'a>),
     Value(Box<Value<'a>>),
 
@@ -115,6 +116,9 @@ impl Hash for Value<'_> {
             Self::U32(inner) => inner.hash(state),
             Self::I64(inner) => inner.hash(state),
             Self::U64(inner) => inner.hash(state),
+            // To hold the +0.0 == -0.0 => hash(+0.0) == hash(-0.0) property.
+            // See https://doc.rust-lang.org/beta/std/hash/trait.Hash.html#hash-and-eq
+            Self::F64(inner) if *inner == 0. => 0f64.to_le_bytes().hash(state),
             Self::F64(inner) => inner.to_le_bytes().hash(state),
             Self::Str(inner) => inner.hash(state),
             Self::Signature(inner) => inner.hash(state),
@@ -139,10 +143,10 @@ impl Ord for Value<'_> {
             .unwrap_or_else(|| match (self, other) {
                 (Self::F64(lhs), Self::F64(rhs)) => lhs.total_cmp(rhs),
                 // `partial_cmp` returns `Some(_)` if either the discriminants are different
-                // or if both the left hand side and right hand side is `Self::F64(_)`,
-                // because `f64` is the only type in this enum, that does not implement `Ord`.
-                // This `match`-arm is therefore unreachable.
-                _ => unreachable!(),
+                // or if both the left hand side and right hand side is `Self::F64(_)`. We can only
+                // reach this arm, if only one of the sides is `Self::F64(_)`. So we can just
+                // pretend the ordering is equal.
+                _ => Ordering::Equal,
             })
     }
 }
@@ -203,7 +207,7 @@ impl<'a> Value<'a> {
         T: Into<Self> + DynamicType,
     {
         // With specialization, we wouldn't have this
-        if value.dynamic_signature() == VARIANT_SIGNATURE_STR {
+        if value.signature() == VARIANT_SIGNATURE_STR {
             Self::Value(Box::new(value.into()))
         } else {
             value.into()
@@ -246,31 +250,31 @@ impl<'a> Value<'a> {
     }
 
     /// Get the signature of the enclosed value.
-    pub fn value_signature(&self) -> Signature<'_> {
+    pub fn value_signature(&self) -> &Signature {
         match self {
-            Value::U8(_) => u8::signature(),
-            Value::Bool(_) => bool::signature(),
-            Value::I16(_) => i16::signature(),
-            Value::U16(_) => u16::signature(),
-            Value::I32(_) => i32::signature(),
-            Value::U32(_) => u32::signature(),
-            Value::I64(_) => i64::signature(),
-            Value::U64(_) => u64::signature(),
-            Value::F64(_) => f64::signature(),
-            Value::Str(_) => <&str>::signature(),
-            Value::Signature(_) => Signature::signature(),
-            Value::ObjectPath(_) => ObjectPath::signature(),
-            Value::Value(_) => Signature::from_static_str_unchecked("v"),
+            Value::U8(_) => u8::SIGNATURE,
+            Value::Bool(_) => bool::SIGNATURE,
+            Value::I16(_) => i16::SIGNATURE,
+            Value::U16(_) => u16::SIGNATURE,
+            Value::I32(_) => i32::SIGNATURE,
+            Value::U32(_) => u32::SIGNATURE,
+            Value::I64(_) => i64::SIGNATURE,
+            Value::U64(_) => u64::SIGNATURE,
+            Value::F64(_) => f64::SIGNATURE,
+            Value::Str(_) => <&str>::SIGNATURE,
+            Value::Signature(_) => Signature::SIGNATURE,
+            Value::ObjectPath(_) => ObjectPath::SIGNATURE,
+            Value::Value(_) => &Signature::Variant,
 
             // Container types
-            Value::Array(value) => value.full_signature().as_ref(),
-            Value::Dict(value) => value.full_signature().as_ref(),
-            Value::Structure(value) => value.full_signature().as_ref(),
+            Value::Array(value) => value.signature(),
+            Value::Dict(value) => value.signature(),
+            Value::Structure(value) => value.signature(),
             #[cfg(feature = "gvariant")]
-            Value::Maybe(value) => value.full_signature().as_ref(),
+            Value::Maybe(value) => value.signature(),
 
             #[cfg(unix)]
-            Value::Fd(_) => Fd::signature(),
+            Value::Fd(_) => Fd::SIGNATURE,
         }
     }
 
@@ -338,6 +342,23 @@ impl<'a> Value<'a> {
         serialize_value!(self serializer.serialize_element)
     }
 
+    pub(crate) fn serialize_value_as_dict_key<S>(&self, serializer: &mut S) -> Result<(), S::Error>
+    where
+        S: SerializeMap,
+    {
+        serialize_value!(self serializer.serialize_key)
+    }
+
+    pub(crate) fn serialize_value_as_dict_value<S>(
+        &self,
+        serializer: &mut S,
+    ) -> Result<(), S::Error>
+    where
+        S: SerializeMap,
+    {
+        serialize_value!(self serializer.serialize_value)
+    }
+
     #[cfg(feature = "gvariant")]
     pub(crate) fn serialize_value_as_some<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
@@ -387,7 +408,7 @@ impl<'a> Value<'a> {
     /// [`Value::Value`]: enum.Value.html#variant.Value
     /// [`TryFrom<Value>`]: https://doc.rust-lang.org/std/convert/trait.TryFrom.html
     /// [`From<Value>`]: https://doc.rust-lang.org/std/convert/trait.From.html
-    pub fn downcast<T: ?Sized>(self) -> Result<T, crate::Error>
+    pub fn downcast<T>(self) -> Result<T, crate::Error>
     where
         T: TryFrom<Value<'a>>,
         <T as TryFrom<Value<'a>>>::Error: Into<crate::Error>,
@@ -439,7 +460,7 @@ impl<'a> Value<'a> {
     /// [`downcast`]: enum.Value.html#method.downcast
     pub fn downcast_ref<T>(&'a self) -> Result<T, crate::Error>
     where
-        T: ?Sized + TryFrom<&'a Value<'a>>,
+        T: TryFrom<&'a Value<'a>>,
         <T as TryFrom<&'a Value<'a>>>::Error: Into<crate::Error>,
     {
         if let Value::Value(v) = self {
@@ -522,7 +543,7 @@ pub(crate) fn value_display_fmt(
             if type_annotate {
                 f.write_str("signature ")?;
             }
-            write!(f, "{:?}", val.as_str())
+            write!(f, "{:?}", val.to_string())
         }
         Value::ObjectPath(val) => {
             if type_annotate {
@@ -555,18 +576,18 @@ pub(crate) fn value_display_fmt(
     }
 }
 
-impl<'a> Serialize for Value<'a> {
+impl Serialize for Value<'_> {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
         S: Serializer,
     {
         // Serializer implementation needs to ensure padding isn't added for Value.
-        let mut structure = serializer.serialize_struct("zvariant::Value", 2)?;
+        let mut structure = serializer.serialize_struct("Variant", 2)?;
 
         let signature = self.value_signature();
-        structure.serialize_field("zvariant::Value::Signature", &signature)?;
+        structure.serialize_field("signature", &signature)?;
 
-        self.serialize_value_as_struct_field("zvariant::Value::Value", &mut structure)?;
+        self.serialize_value_as_struct_field("value", &mut structure)?;
 
         structure.end()
     }
@@ -600,11 +621,11 @@ impl<'de> Visitor<'de> for ValueVisitor {
     where
         V: SeqAccess<'de>,
     {
-        let signature = visitor.next_element::<Signature<'_>>()?.ok_or_else(|| {
+        let signature = visitor.next_element::<Signature>()?.ok_or_else(|| {
             Error::invalid_value(Unexpected::Other("nothing"), &"a Value signature")
         })?;
         let seed = ValueSeed::<Value<'_>> {
-            signature,
+            signature: &signature,
             phantom: PhantomData,
         };
 
@@ -617,35 +638,41 @@ impl<'de> Visitor<'de> for ValueVisitor {
     where
         V: MapAccess<'de>,
     {
-        let (_, signature) = visitor
-            .next_entry::<&str, Signature<'_>>()?
-            .ok_or_else(|| {
-                Error::invalid_value(Unexpected::Other("nothing"), &"a Value signature")
-            })?;
+        let (_, signature) = visitor.next_entry::<&str, Signature>()?.ok_or_else(|| {
+            Error::invalid_value(Unexpected::Other("nothing"), &"a Value signature")
+        })?;
         let _ = visitor.next_key::<&str>()?;
 
         let seed = ValueSeed::<Value<'_>> {
-            signature,
+            signature: &signature,
             phantom: PhantomData,
         };
         visitor.next_value_seed(seed)
     }
 }
 
-pub(crate) struct SignatureSeed<'de> {
-    pub signature: Signature<'de>,
+pub(crate) struct SignatureSeed<'sig> {
+    pub signature: &'sig Signature,
 }
 
-impl<'de> SignatureSeed<'de> {
-    pub(crate) fn visit_array<V>(self, mut visitor: V) -> Result<Array<'de>, V::Error>
+impl SignatureSeed<'_> {
+    pub(crate) fn visit_array<'de, V>(self, mut visitor: V) -> Result<Array<'de>, V::Error>
     where
         V: SeqAccess<'de>,
     {
-        let element_signature = self.signature.slice(1..);
-        let mut array = Array::new_full_signature(self.signature.clone());
+        let element_signature = match self.signature {
+            Signature::Array(child) => child.signature(),
+            _ => {
+                return Err(Error::invalid_type(
+                    Unexpected::Str(&self.signature.to_string()),
+                    &"an array signature",
+                ))
+            }
+        };
+        let mut array = Array::new_full_signature(self.signature);
 
         while let Some(elem) = visitor.next_element_seed(ValueSeed::<Value<'_>> {
-            signature: element_signature.clone(),
+            signature: element_signature,
             phantom: PhantomData,
         })? {
             elem.value_signature();
@@ -655,20 +682,22 @@ impl<'de> SignatureSeed<'de> {
         Ok(array)
     }
 
-    pub(crate) fn visit_struct<V>(self, mut visitor: V) -> Result<Structure<'de>, V::Error>
+    pub(crate) fn visit_struct<'de, V>(self, mut visitor: V) -> Result<Structure<'de>, V::Error>
     where
         V: SeqAccess<'de>,
     {
-        let mut i = 1;
-        let signature_end = self.signature.len() - 1;
-        let mut builder = StructureBuilder::new();
-        while i < signature_end {
-            let fields_signature = self.signature.slice(i..signature_end);
-            let parser = SignatureParser::new(fields_signature.as_ref());
-            let len = parser.next_signature().map_err(Error::custom)?.len();
-            let field_signature = fields_signature.slice(0..len);
-            i += field_signature.len();
+        let fields_signatures = match self.signature {
+            Signature::Structure(fields) => fields.iter(),
+            _ => {
+                return Err(Error::invalid_type(
+                    Unexpected::Str(&self.signature.to_string()),
+                    &"a structure signature",
+                ))
+            }
+        };
 
+        let mut builder = StructureBuilder::new();
+        for field_signature in fields_signatures {
             if let Some(field) = visitor.next_element_seed(ValueSeed::<Value<'_>> {
                 signature: field_signature,
                 phantom: PhantomData,
@@ -680,20 +709,20 @@ impl<'de> SignatureSeed<'de> {
     }
 }
 
-impl<'de, T> From<ValueSeed<'de, T>> for SignatureSeed<'de> {
-    fn from(seed: ValueSeed<'de, T>) -> Self {
+impl<'sig, T> From<ValueSeed<'sig, T>> for SignatureSeed<'sig> {
+    fn from(seed: ValueSeed<'sig, T>) -> Self {
         SignatureSeed {
             signature: seed.signature,
         }
     }
 }
 
-struct ValueSeed<'de, T> {
-    signature: Signature<'de>,
+struct ValueSeed<'sig, T> {
+    signature: &'sig Signature,
     phantom: PhantomData<T>,
 }
 
-impl<'de, T> ValueSeed<'de, T>
+impl<'de, T> ValueSeed<'_, T>
 where
     T: Deserialize<'de>,
 {
@@ -718,7 +747,7 @@ where
     }
 
     #[inline]
-    fn visit_variant<V>(self, visitor: V) -> Result<Value<'de>, V::Error>
+    fn visit_variant_as_seq<V>(self, visitor: V) -> Result<Value<'de>, V::Error>
     where
         V: SeqAccess<'de>,
     {
@@ -726,12 +755,22 @@ where
             .visit_seq(visitor)
             .map(|v| Value::Value(Box::new(v)))
     }
+
+    #[inline]
+    fn visit_variant_as_map<V>(self, visitor: V) -> Result<Value<'de>, V::Error>
+    where
+        V: MapAccess<'de>,
+    {
+        ValueVisitor
+            .visit_map(visitor)
+            .map(|v| Value::Value(Box::new(v)))
+    }
 }
 
 macro_rules! value_seed_basic_method {
     ($name:ident, $type:ty) => {
         #[inline]
-        fn $name<E>(self, value: $type) -> Result<Value<'de>, E>
+        fn $name<E>(self, value: $type) -> Result<Value<'static>, E>
         where
             E: serde::de::Error,
         {
@@ -740,34 +779,7 @@ macro_rules! value_seed_basic_method {
     };
 }
 
-macro_rules! value_seed_str_method {
-    ($name:ident, $type:ty, $constructor:ident) => {
-        fn $name<E>(self, value: $type) -> Result<Value<'de>, E>
-        where
-            E: serde::de::Error,
-        {
-            match self.signature.as_str() {
-                <&str>::SIGNATURE_STR => Ok(Value::Str(Str::from(value))),
-                Signature::SIGNATURE_STR => Ok(Value::Signature(Signature::$constructor(value))),
-                ObjectPath::SIGNATURE_STR => Ok(Value::ObjectPath(ObjectPath::$constructor(value))),
-                _ => {
-                    let expected = format!(
-                        "`{}`, `{}` or `{}`",
-                        <&str>::SIGNATURE_STR,
-                        Signature::SIGNATURE_STR,
-                        ObjectPath::SIGNATURE_STR,
-                    );
-                    Err(Error::invalid_type(
-                        Unexpected::Str(self.signature.as_str()),
-                        &expected.as_str(),
-                    ))
-                }
-            }
-        }
-    };
-}
-
-impl<'de, T> Visitor<'de> for ValueSeed<'de, T>
+impl<'de, T> Visitor<'de> for ValueSeed<'_, T>
 where
     T: Deserialize<'de>,
 {
@@ -790,14 +802,9 @@ where
     where
         E: serde::de::Error,
     {
-        let v = match self.signature.as_bytes().first().ok_or_else(|| {
-            Error::invalid_value(
-                Unexpected::Other("nothing"),
-                &"i32 or fd signature character",
-            )
-        })? {
+        let v = match &self.signature {
             #[cfg(unix)]
-            b'h' => {
+            Signature::Fd => {
                 // SAFETY: The `'de` lifetimes will ensure the borrow won't outlive the raw FD.
                 let fd = unsafe { std::os::fd::BorrowedFd::borrow_raw(value) };
                 Fd::Borrowed(fd).into()
@@ -816,24 +823,42 @@ where
         self.visit_string(String::from(value))
     }
 
-    value_seed_str_method!(visit_borrowed_str, &'de str, from_str_unchecked);
+    fn visit_borrowed_str<E>(self, v: &'de str) -> Result<Self::Value, E>
+    where
+        E: Error,
+    {
+        match &self.signature {
+            Signature::Str => Ok(Value::Str(Str::from(v))),
+            Signature::Signature => Signature::try_from(v)
+                .map(Value::Signature)
+                .map_err(Error::custom),
+            Signature::ObjectPath => Ok(Value::ObjectPath(ObjectPath::from_str_unchecked(v))),
+            _ => {
+                let expected = format!(
+                    "`{}`, `{}` or `{}`",
+                    <&str>::SIGNATURE_STR,
+                    Signature::SIGNATURE_STR,
+                    ObjectPath::SIGNATURE_STR,
+                );
+                Err(Error::invalid_type(
+                    Unexpected::Str(&self.signature.to_string()),
+                    &expected.as_str(),
+                ))
+            }
+        }
+    }
 
     fn visit_seq<V>(self, visitor: V) -> Result<Value<'de>, V::Error>
     where
         V: SeqAccess<'de>,
     {
-        match self.signature.as_bytes().first().ok_or_else(|| {
-            Error::invalid_value(
-                Unexpected::Other("nothing"),
-                &"Array or Struct signature character",
-            )
-        })? {
+        match &self.signature {
             // For some reason rustc doesn't like us using ARRAY_SIGNATURE_CHAR const
-            b'a' => self.visit_array(visitor),
-            b'(' => self.visit_struct(visitor),
-            b'v' => self.visit_variant(visitor),
-            b => Err(Error::invalid_value(
-                Unexpected::Char(*b as char),
+            Signature::Array(_) => self.visit_array(visitor),
+            Signature::Structure(_) => self.visit_struct(visitor),
+            Signature::Variant => self.visit_variant_as_seq(visitor),
+            s => Err(Error::invalid_value(
+                Unexpected::Str(&s.to_string()),
                 &"a Value signature",
             )),
         }
@@ -843,24 +868,26 @@ where
     where
         V: MapAccess<'de>,
     {
-        if self.signature.len() < 5 {
-            return Err(serde::de::Error::invalid_length(
-                self.signature.len(),
-                &">= 5 characters in dict entry signature",
-            ));
-        }
-        let key_signature = self.signature.slice(2..3);
-        let signature_end = self.signature.len() - 1;
-        let value_signature = self.signature.slice(3..signature_end);
-        let mut dict = Dict::new_full_signature(self.signature.clone());
+        let (key_signature, value_signature) = match &self.signature {
+            Signature::Dict { key, value } => (key.signature().clone(), value.signature().clone()),
+            Signature::Variant => return self.visit_variant_as_map(visitor),
+            _ => {
+                return Err(Error::invalid_type(
+                    Unexpected::Str(&self.signature.to_string()),
+                    &"a dict signature",
+                ))
+            }
+        };
+
+        let mut dict = Dict::new_full_signature(self.signature);
 
         while let Some((key, value)) = visitor.next_entry_seed(
             ValueSeed::<Value<'_>> {
-                signature: key_signature.clone(),
+                signature: &key_signature,
                 phantom: PhantomData,
             },
             ValueSeed::<Value<'_>> {
-                signature: value_signature.clone(),
+                signature: &value_signature,
                 phantom: PhantomData,
             },
         )? {
@@ -875,8 +902,17 @@ where
     where
         D: Deserializer<'de>,
     {
+        let child_signature = match &self.signature {
+            Signature::Maybe(child) => child.signature().clone(),
+            _ => {
+                return Err(Error::invalid_type(
+                    Unexpected::Str(&self.signature.to_string()),
+                    &"a maybe signature",
+                ))
+            }
+        };
         let visitor = ValueSeed::<T> {
-            signature: self.signature.slice(1..),
+            signature: &child_signature,
             phantom: PhantomData,
         };
 
@@ -912,7 +948,7 @@ where
     }
 }
 
-impl<'de, T> DeserializeSeed<'de> for ValueSeed<'de, T>
+impl<'de, T> DeserializeSeed<'de> for ValueSeed<'_, T>
 where
     T: Deserialize<'de>,
 {
@@ -926,10 +962,8 @@ where
     }
 }
 
-impl<'a> Type for Value<'a> {
-    fn signature() -> Signature<'static> {
-        Signature::from_static_str_unchecked(VARIANT_SIGNATURE_STR)
-    }
+impl Type for Value<'_> {
+    const SIGNATURE: &'static Signature = &Signature::Variant;
 }
 
 impl<'a> TryFrom<&Value<'a>> for Value<'a> {
@@ -940,6 +974,20 @@ impl<'a> TryFrom<&Value<'a>> for Value<'a> {
     }
 }
 
+impl Clone for Value<'_> {
+    /// Clone the value.
+    ///
+    /// # Panics
+    ///
+    /// This method can only fail on Unix platforms for [`Value::Fd`] variant containing an
+    /// [`Fd::Owned`] variant. This happens when the current process exceeds the limit on maximum
+    /// number of open file descriptors.
+    fn clone(&self) -> Self {
+        self.try_clone()
+            .expect("Process exceeded limit on maximum number of open file descriptors")
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use std::collections::HashMap;
@@ -986,8 +1034,8 @@ mod tests {
         assert_eq!(
             Value::new((
                 vec![
-                    Signature::from_static_str("").unwrap(),
-                    Signature::from_static_str("(ysa{sd})").unwrap(),
+                    Signature::try_from("").unwrap(),
+                    Signature::try_from("(ysa{sd})").unwrap(),
                 ],
                 vec![
                     ObjectPath::from_static_str("/").unwrap(),
@@ -1063,7 +1111,6 @@ mod tests {
             .iter()
             .any(|str| str.contains("100") && str.contains(": ") && str.contains("200")));
 
-        assert_eq!(Value::new(Structure::default()).to_string(), "()");
         assert_eq!(
             Value::new(((true,), (true, false), (true, true, false))).to_string(),
             "((true,), (true, false), (true, true, false))"
